import numpy as np
from .splash import welcome


def generate2dkmesh(x1, y1, x2, y2, z, nkx, nky):
    """
	This module generates a KPOINTS mesh file for 2D plotting.
	"""
    welcome()

    kx = np.linspace(x1, x2, nkx)
    ky = np.linspace(y1, y2, nky)
    wf = open("Kgrid.dat", "w")
    wf.write("Generated by PyProcar\n")
    wf.write("%d\n" % (nkx * nky))
    wf.write("Reciprocal\n")
    for ikx in kx:
        for iky in ky:
            wf.write(
                " {: >12.7f}   {: >12.7f}   {: >12.7f}   {: >12.7f}\n".format(
                    ikx, iky, z, 1.0
                )
            )
    wf.close()



import numpy as np
import re
import logging
import matplotlib.pyplot as plt
import sys
from ..utilsprocar import UtilsProcar


class ProcarFileFilter:
    """Process a PROCAR file fields line-wise, specially useful for HUGE
  files. This could be thought as pre-processing, writting a new
  PROCAR-like file but reduced in some way.

  A PROCAR File is basically an multi-dimmensional arrays of data, the
  largest being:
  spd_data[#kpoints][#band][#ispin][#atom][#orbital]
  
  while the number of Kpoints d'ont seems a target for reduction
  (omission or averaging), the other fields can be reduced, for
  instance: grouping the atoms by species or as "surtrate" and
  "adsorbate", or just keeping the bands close to the Fermi energy, or
  discarding the d-orbitals in a s-p system. You got the idea, rigth?

  Example:

  -To group the "s", "p" y "d" orbitals from the file PROCAR and write
   them in PROCAR-spd:
   
   >>> a = procar.ProcarFileFilter("PROCAR", "PROCAR-new")
   >>> a.FilterOrbitals([[0],[1,2,3],[4,5,6,7,8]], ['s','p', 'd'])
   
       The PROCAR-new will have just 3+1 columns (orbitals are colum-wise
       , take a look to the file). If you omit the ['s', 'p', 'd'] list, 
       the new orbitals will have a generic meaningless name (o1, o2, o3)

  -To group the atoms 1,2,5,6 and 3,4,7,8 from PROCAR and write them
   in PROCAR-new (note the 0-based indexes):

   >>> a = procar.ProcarFileFilter("PROCAR", "PROCAR-new")
   >>> a.FilterAtoms([[0,1,4,5],[2,3,6,7]])

   -To select just the total density (ie: ignoring the spin-resolved stuff, 
    if any)from PROCAR and write it in PROCAR-new:

   >>> a = procar.ProcarFileFilter("PROCAR", "PROCAR-new")
   >>> a.FilterSpin([0])

  """

    def __init__(self, infile=None, outfile=None, loglevel=logging.WARNING):
        """Initialize the class.

    Params: `infile=None`, input fileName
    """
        self.infile = infile
        self.outfile = outfile

        # We want a logging to tell us what is happening
        self.log = logging.getLogger("ProcarFileFilter")
        self.log.setLevel(loglevel)
        # This is a handler for logging, by now just keep it
        # untouched. Dont really matters its usage
        self.ch = logging.StreamHandler()
        self.ch.setFormatter(
            logging.Formatter("%(name)s::%(levelname)s:" " %(message)s")
        )
        self.ch.setLevel(logging.DEBUG)
        self.log.addHandler(self.ch)
        # At last, one message to the logger.
        self.log.debug("ProcarFileFilter instanciated")
        return

    ##########################SCRIPTS#################################

    # def scriptFilter(self,inFile,outFile,atoms=None,orbitals=None,orbital_names=None,bands=None,spin=None,human_atoms=False):
    #   print "Input file  :", inFile
    #   print "Output file :", outFile

    #   print "atoms       :", atoms
    #   if atoms:
    #     print "human_atoms     :", human_atoms
    #   print "orbitals  :", orbitals
    #   if orbitals:
    #       print "orb. names  :", orbital_names
    #     print "bands       :", bands
    #     print "spins       :", spin

    #   #Access init class of ProcarFileFilter and pass two arguments
    #   FileFilter = ProcarFileFilter(inFile,outFile)

    #   #for atoms
    #   if atoms:
    #     print "Manipulating the atoms"

    #     if human_atoms:
    #       atoms = [[y-1 for y in x] for x in atoms]
    #       print "new atoms list :", atoms

    #     #Now just left to call the driver member
    #     FileFilter.FilterAtoms(atoms)

    #   #for orbitals
    #   elif orbitals:
    #     print "Manipulating the orbitals"
    #     #If orbitals orbital_names is None, it needs to be filled
    #     if orbital_names is None:
    #       orbital_names = ["o"+str(x) for x in range(len(orbitals))]
    #       print "New orbitals names (default): ", orbital_names
    #     #testing if makes sense
    #     if len(orbitals) != len(orbital_names):
    #       raise RuntimeError("length of orbitals and orbitals names do not match")

    #     FileFilter.FilterOrbitals(orbitals,orbital_names)

    #   #for bands
    #   elif bands:
    #     print "Manipulating the bands"

    #     bmin = bands[0]
    #     bmax = bands[1]
    #     if bmax < bmin:
    #       bmax, bmin = bmin, bmax
    #       print "New bands limits: ", bmin, " to ", bmax

    #     FileFilter.FilterBands(bmin,bmax)

    #   #for spin
    #   elif spin:
    #     print "Manipulating the spin"

    #     FileFilter.FilterSpin(spin)

    #   return

    ##################################################################

    def setInFile(self, infile):
        """Sets a input file `infile`, it can contains the path to the file"""
        self.infile = infile
        self.log.info("Input File: " + infile)
        return

    def setOutFile(self, outfile):
        """Sets a output file `outfile`, it can contains the path to the file"""
        self.outfile = outfile
        self.log.info("Out File: " + outfile)
        return

    def FilterOrbitals(self, orbitals, orbitalsNames):
        """
    Reads the file already set by SetInFile() and writes a new
    file already set by SetOutFile(). The new file only has the
    selected/grouped orbitals.

    Args: 

    -orbitals: nested iterable with the orbitals indexes to be
      considered. For example: [[0],[2]] means select the first
      orbital ("s") and the second one ("pz").
      [[0],[1,2,3],[4,5,6,7,8]] is ["s", "p", "d"].

    -orbitalsNames: The name to be put in each new orbital field (of a
      orbital line). For example ["s","p","d"] is a good
      `orbitalsName` for the `orbitals`=[[0],[1,2,3],[4,5,6,7,8]]. 
      However, ["foo", "bar", "baz"] is equally valid.

    Note: 
      -The atom index is not counted as the first field.
      -The last column ('tot') is so important that it is always
       included. Do not needs to be called
    """
        # setting iostuff, this method -and class- should not made any
        # checking about IO, that is the job of the caller
        self.log.info("In File: " + self.infile)
        self.log.info("Out File: " + self.outfile)
        # open the files
        fout = open(self.outfile, "w")
        fopener = UtilsProcar()
        fin = fopener.OpenFile(self.infile)
        for line in fin:
            if re.match(r"\s*ion\s*", line):
                # self.log.debug("orbital line found: " + line)
                line = " ".join(["ion"] + orbitalsNames + ["tot"]) + "\n"

            elif re.match(r"\s*\d+\s*", line) or re.match(r"\s*tot\s*", line):
                # self.log.debug("data line found: " + line)
                line = line.split()
                # all floats to an array
                data = np.array(line[1:], dtype=float)
                # setting a new line, keeping just the first value
                line = line[:1]
                for orbset in orbitals:
                    line.append(data[orbset].sum())
                # the last value ("tot") always  should be written
                line.append(data[-1])
                # converting to str
                line = [str(x) for x in line]
                line = " ".join(line) + "\n"
            fout.write(line)

        return

    def FilterAtoms(self, atomsGroups):
        """
    Reads the file already set by SetInFile() and writes a new
    file already set by SetOutFile(). The new file only has the
    selected/grouped atoms.

    Args: 

    -atomsGroups: nested iterable with the atoms indexes (0-based) to
      be considered. For example: [[0],[2]] means select the first and
      the second atoms. While [[1,2,3],[4,5,6,7,8]] means select the
      contribution of atoms 1+2+3 and 4+5+6+7+8

    Note: 
      -The atom index is c-based (or python) beginning with 0
      -The output has a dummy atom index, without any intrisic meaning
    
    """
        # setting iostuff, this method -and class- should not made any
        # checking about IO, that is the job of the caller
        self.log.info("In File: " + self.infile)
        self.log.info("Out File: " + self.outfile)
        # open the files
        fout = open(self.outfile, "w")
        fopener = UtilsProcar()
        with fopener.OpenFile(self.infile) as fin:
            # I need to change the numbers of ions, it will needs the second
            # line. The first one is not needed
            fout.write(fin.readline())
            line = fin.readline()
            line = line.split()
            # the very last value needs to be changed
            line[-1] = str(len(atomsGroups))
            line = " ".join(line)
            fout.write(line + "\n")

            # now parsing the rest of the file
            data = []
            for line in fin:
                # if line has data just capture it
                if re.match(r"\s*\d+\s*", line):
                    # self.log.debug("atoms line found: " + line)
                    data.append(line)
                # if `line` is a end of th block (begins with 'tot'), do the
                # work. And clean up data then
                elif re.match(r"\s*tot\s*", line):
                    # self.log.debug("tot line found: " + line)
                    # making an array
                    data = [x.split() for x in data]
                    data = np.array(data, dtype=float)
                    # iterating on the atoms groups
                    for index in range(len(atomsGroups)):
                        atoms = atomsGroups[index]
                        # summing colum-wise
                        atomLine = data[atoms].sum(axis=0)
                        atomLine = [str(x) for x in atomLine]
                        # the atom index should not be averaged (anyway now is
                        # meaningless)
                        atomLine[0] = str(index + 1)
                        atomLine = " ".join(atomLine)
                        fout.write(atomLine + "\n")

                    # clean the buffer
                    data = []
                    # and write the `tot` line
                    fout.write(line)
                # otherwise just write this line
                else:
                    fout.write(line)

        return

    def FilterBands(self, Min, Max):
        """
    Reads the file already set by SetInFile() and writes a new
    file already set by SetOutFile(). The new file only has the
    selected bands.

    Args: 

    -Min, Max:
      the minimum/maximum band  index to be considered, the indexes 
      are the same used by vasp (ie written in the file).


    Note: -Since bands are somewhat disordered in vasp you may like to
      consider a large region and made some trial and error
    
    """
        # setting iostuff, this method -and class- should not made any
        # checking about IO, that is the job of the caller
        self.log.info("In File: " + self.infile)
        self.log.info("Out File: " + self.outfile)
        # open the files
        fout = open(self.outfile, "w")
        fopener = UtilsProcar()
        fin = fopener.OpenFile(self.infile)

        # I need to change the numbers of kpoints, it will needs the second
        # line. The first one is not needed
        fout.write(fin.readline())
        line = fin.readline()
        # the third value needs to be changed, however better print it
        self.log.debug("The line contaning bands number is " + line)
        line = line.split()
        self.log.debug("The number of bands is: " + line[7])
        line[7] = str(Max - Min + 1)
        line = " ".join(line)
        fout.write(line + "\n")

        # now parsing the rest of the file
        write = True
        for line in fin:
            if re.match(r"\s*band\s*", line):
                # self.log.debug("bands line found: " + line)
                band = int(re.match(r"\s*band\s*(\d+)", line).group(1))
                if band < Min or band > Max:
                    write = False
                else:
                    write = True
            if re.match(r"\s*k-point\s*", line):
                write = True
            if write:
                fout.write(line)
        return

    def FilterSpin(self, components):
        """Reads the file already set by SetInFile() and writes a new
    file already set by SetOutFile(). The new file only has the
    selected part of the density (sigma_i).

    Args: 

    -components: The spin component block, for instante [0] menas just
      the density, while [1,2] would be the the sigma_x and sigma_y
      for a non-collinear calculation.

    TODO: spin-polarized collinear case is not included at all, not
    even a warning message!

    """
        # setting iostuff, this method -and class- should not made any
        # checking about IO, that is the job of the caller
        self.log.info("In File: " + self.infile)
        self.log.info("Out File: " + self.outfile)
        # open the files
        fout = open(self.outfile, "w")
        fopener = UtilsProcar()
        with fopener.OpenFile(self.infile) as fin:
            counter = 0
            for line in fin:
                # if any data found
                if re.match(r"\s*\d", line):
                    # check if should be written
                    if counter in components:
                        fout.write(line)
                elif re.match(r"\s*tot", line):
                    if counter in components:
                        fout.write(line)
                    # the next block will belong to other component
                    counter += 1
                elif re.match(r"\s*ion", line):
                    fout.write(line)
                    counter = 0
                else:
                    fout.write(line)
        return

    def FilterKpoints(self, Min, Max):
        """
    Reads the file already set by SetInFile() and writes a new
    file already set by SetOutFile(). The new file the
    selected bands.

    Args: 

    -Min, Max:
      the minimum/maximum band  kpoint to be considered, the indexes 
      are the same used by vasp (i.e. written in the file). Not starting from zero
    """
        # setting iostuff, this method -and class- should not made any
        # checking about IO, that is the job of the caller
        self.log.info("In File: " + self.infile)
        self.log.info("Out File: " + self.outfile)
        # open the files
        fout = open(self.outfile, "w")
        fopener = UtilsProcar()
        fin = fopener.OpenFile(self.infile)

        # I need to change the numbers of kpoints, it will needs the second
        # line. The first one is not needed
        fout.write(fin.readline())
        line = fin.readline()
        # the third value needs to be changed, however better print it
        self.log.debug("The line contaning kpoints number is " + line)
        line = line.split()
        self.log.debug("The number of kpoints is: " + line[3])
        line[3] = str(Max - Min + 1)
        line = " ".join(line)
        fout.write(line + "\n")

        # now parsing the rest of the file
        write = True
        for line in fin:
            if re.match(r"\s*k-point\s*", line):
                self.log.debug("bands line found: " + line)
                kpoint = int(re.match(r"\s*k-point\s*(\d+)", line).group(1))
                if kpoint < Min or kpoint > Max:
                    write = False
                else:
                    write = True
            if write:
                fout.write(line)
        return



"""
Created on Fri March 2 2020
@author: Pedram Tavadze

"""

from re import findall
from numpy import array, linspace, where, zeros, dot, sum


class ElkParser:
    def __init__(self, elkin="elk.in", kdirect=True):

        # elk specific inp
        self.fin = elkin
        self.file_names = []
        self.high_symmetry_points = None
        self.nhigh_sym = None
        self.nspecies = None
        self.nbands = None
        self.composition = {}
        self.tasks = None
        self.kticks = None
        self.knames = None
        self.kdirect = kdirect

        self.kpoints = None
        self.kpointsCount = None
        self.bands = None
        self.bandsCount = None
        self.ionsCount = None

        self.spd = None
        self.cspd = None

        self.fermi = None
        self.reclat = None  # reciprocal lattice vectors

        self.orbitalName = [
            "Y00",
            "Y1-1",
            "Y10",
            "Y11",
            "Y2-2",
            "Y2-1",
            "Y20",
            "Y21",
            "Y22",
            "Y3-3",
            "Y3-2",
            "Y3-1",
            "Y30",
            "Y3-1",
            "Y3-2",
            "Y3-3",
        ]
        self.orbitalName_short = ["s", "p", "d", "f"]
        self.orbitalCount = None  # number of orbitals

        # number of spin components (blocks of data), 1: non-magnetic non
        # polarized, 2: spin polarized collinear, 4: non-collinear
        # spin.
        # NOTE: before calling to `self._readOrbital` the case '4'
        # is marked as '1'
        self.ispin = None

        rf = open(self.fin, "r")
        self.elkIn = rf.read()
        rf.close()

        self._readElkin()
        self._readFiles()

        self._readFermi()
        self._readRecLattice()

        return

    @property
    def nspin(self):
        """
        number of spin, default is 1.
        """
        nspindict = {1: 1, 2: 2, 4: 2, None: 1}
        return nspindict[self.ispin]

    @property
    def spd_orb(self):
        # indices: ikpt, iband, ispin, iion, iorb
        # remove indices and total from iorb.
        return self.spd[:, :, :, 1:-1]

    def _readElkin(self):
        self.nhigh_sym, self.kpointsCount = [
            int(x) for x in findall("plot1d\n\s*([0-9]*)\s*([0-9]*)", self.elkIn)[0]
        ]
        raw_hsp = findall(
            "plot1d\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * "([0-9\s\.-]*).*\n",
            self.elkIn,
        )[0]
        self.high_symmetry_points = zeros(shape=(self.nhigh_sym, 3))
        for ihs in range(self.nhigh_sym):
            self.high_symmetry_points[ihs, :] = [
                float(x) for x in raw_hsp[ihs].split()[0:3]
            ]
        self.nspecies = int(findall("atoms\n\s*([0-9]*)", self.elkIn)[0])
        self.ionsCount = 0
        for ispc in findall("'([A-Za-z]*).in'.*\n\s*([0-9]*)", self.elkIn):
            self.composition[ispc[0]] = int(ispc[1])
            self.ionsCount += int(ispc[1])
        raw_ticks = findall(
            "plot1d\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * ".*:(.*)\n", self.elkIn
        )[0]
        if len(raw_ticks) != self.nhigh_sym:
            self.knames = [str(x) for x in range(self.nhigh_sym)]
        else:
            self.knames = [
                "$%s$" % (x.replace(",", "").replace("vlvp1d", "").replace(" ", ""))
                for x in findall(
                    "plot1d\n\s*[0-9]*\s*[0-9]*.*\n" + self.nhigh_sym * ".*:(.*)\n",
                    self.elkIn,
                )[0]
            ]
        self.tasks = [
            int(x) for x in findall("tasks\n\s*([0-9\s\n]*)", self.elkIn)[0].split()
        ]

        if 20 in self.tasks:
            self.file_names.append("BANDS.OUT")
        if 21 in self.tasks or 22 in self.tasks:
            ispc = 1
            for spc in self.composition:
                for iatom in range(self.composition[spc]):
                    self.file_names.append(
                        "BAND_S{:02d}_A{:04d}.OUT".format(ispc, iatom + 1)
                    )
                ispc += 1

    def _readFiles(self):
        rf = open(self.file_names[0], "r")
        lines = rf.readlines()
        rf.close()
        self.bandsCount = int(len(lines) / (self.kpointsCount + 1))
        self.bands = zeros(shape=(self.kpointsCount, self.bandsCount))
        rf = open("BANDLINES.OUT", "r")
        bandLines = rf.readlines()
        rf.close()
        tick_pos = []
        # using strings for a better comparision and avoiding rounding by python
        for iline in range(0, len(bandLines), 3):
            tick_pos.append(bandLines[iline].split()[0])
        x_points = []
        for iline in range(self.kpointsCount):
            x_points.append(lines[iline].split()[0])

        self.kpoints = zeros(shape=(self.kpointsCount, 3))
        x_points = array(x_points)
        tick_pos = array(tick_pos)

        self.kticks = []
        for ihs in range(1, self.nhigh_sym):
            # if ihs == 1:
            #     start = 0
            # else:

            start = where(x_points == tick_pos[ihs - 1])[0][0]
            end = where(x_points == tick_pos[ihs])[0][0] + 1
            self.kpoints[start:end][:] = linspace(
                self.high_symmetry_points[ihs - 1],
                self.high_symmetry_points[ihs],
                end - start,
            )
            self.kticks.append(start)
        self.kticks.append(self.kpointsCount - 1)

        rf = open(self.file_names[0], "r")
        lines = rf.readlines()
        rf.close()

        iline = 0
        for iband in range(self.bandsCount):
            for ikpoint in range(self.kpointsCount):
                self.bands[ikpoint, iband] = float(lines[iline].split()[1])
                iline += 1
            if ikpoint == self.kpointsCount - 1:
                iline += 1
        self.bands *= 27.21138386
        self.spinCount = 1
        self.orbitalCount = 16
        self.spd = zeros(
            shape=(
                self.kpointsCount,
                self.bandsCount,
                self.spinCount,
                self.ionsCount + 1,
                self.orbitalCount + 2,
            )
        )
        idx_bands_out = None
        for ifile in range(len(self.file_names)):
            if self.file_names[ifile] == "BANDS.OUT":
                idx_bands_out = ifile
        if idx_bands_out != None:
            del self.file_names[idx_bands_out]

        for ifile in range(self.ionsCount):
            rf = open(self.file_names[ifile], "r")
            lines = rf.readlines()
            rf.close()
            iline = 0

            for iband in range(self.bandsCount):
                for ikpoint in range(self.kpointsCount):
                    temp = array([float(x) for x in lines[iline].split()])
                    self.spd[ikpoint, iband, 0, ifile, 0] = ifile + 1
                    self.spd[ikpoint, iband, 0, ifile, 1:-1] = temp[2:]
                    iline += 1
                if ikpoint == self.kpointsCount - 1:
                    iline += 1
        # self.spd[:,:,:,-1,:] = self.spd.sum(axis=3)
        self.spd[:, :, :, :, -1] = sum(self.spd[:, :, :, :, 1:-1],axis=4)
        self.spd[:, :, :, -1, :] = self.spd.sum(axis=3)
        self.spd[:, :, 0, -1, 0] = 0

    def _readFermi(self):
        rf = open("EFERMI.OUT", "r")
        self.fermi = float(rf.readline().split()[0])
        rf.close()

    def _readRecLattice(self):
        rf = open("LATTICE.OUT", "r")
        data = rf.read()
        rf.close()
        lattice_block = findall(r"matrix\s*:([\s0-9.]*)Inverse", data)
        lattice_array = array(lattice_block[1].split(), dtype=float)
        self.reclat = lattice_array.reshape((3, 3))

        if self.kdirect is False:
            self.kpoints = dot(self.kpoints, self.reclat)



import numpy as np
import re
import logging
import matplotlib.pyplot as plt
import sys


class ProcarSymmetry:
    def __init__(
        self,
        kpoints,
        bands,
        character=None,
        sx=None,
        sy=None,
        sz=None,
        loglevel=logging.WARNING,
    ):
        self.log = logging.getLogger("ProcarSymmetry")
        self.log.setLevel(loglevel)
        self.ch = logging.StreamHandler()
        self.ch.setFormatter(
            logging.Formatter("%(name)s::%(levelname)s: " "%(message)s")
        )
        self.ch.setLevel(logging.DEBUG)
        self.log.addHandler(self.ch)
        self.log.debug("ProcarSymmetry.__init__: ...")

        self.kpoints = kpoints
        self.bands = bands
        # optional arguments when not given will False, but they can still
        # be treated like arrays
        self.character = np.array([])
        if character is not None:
            self.character = character
        self.sx = np.array([])
        if sx is not None:
            self.sx = sx
        self.sy = np.array([])
        if sy is not None:
            self.sy = sy
        self.sz = np.array([])
        if sz is not None:
            self.sz = sz

        self.log.info("Kpoints : " + str(self.kpoints.shape))
        self.log.info("bands   : " + str(self.bands.shape))
        self.log.info("character  : " + str(self.character.shape))
        self.log.info("sx      : " + str(self.sx.shape))
        self.log.info("sy      : " + str(self.sy.shape))
        self.log.info("sz      : " + str(self.sz.shape))
        self.log.debug("ProcarSymmetry.__init__: ...Done")

        return

    def _q_mult(self, q1, q2):
        """
    Multiplication of quaternions, it doesn't fit in any other place
    """
        w1, x1, y1, z1 = q1
        w2, x2, y2, z2 = q2
        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
        z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
        return np.array((w, x, y, z))

    def GeneralRotation(self, angle, rotAxis=[0, 0, 1], store=True):
        """Apply a rotation defined by an angle and an axis.
    
    Returning value: (Kpoints, sx,sy,sz), the rotated Kpoints and spin
                     vectors (if not the case, they will be empty
                     arrays).

    Arguments
    angle: the rotation angle, must be in degrees!

    rotAxis : a fixed Axis when applying the symmetry, usually it is
    from Gamma to another point). It doesn't need to be normalized. 
    The RotAxis can be:
       [x,y,z] : a cartesian vector in k-space.
       'x': [1,0,0], a rotation in the yz plane. 
       'y': [0,1,0], a rotation in the zx plane.
       'z': [0,0,1], a rotation in the xy plane

    """
        if rotAxis == "x" or rotAxis == "X":
            rotAxis = [1, 0, 0]
        if rotAxis == "y" or rotAxis == "Y":
            rotAxis = [0, 1, 0]
        if rotAxis == "z" or rotAxis == "Z":
            rotAxis = [0, 0, 1]
        rotAxis = np.array(rotAxis, dtype=float)
        self.log.debug("rotAxis : " + str(rotAxis))
        rotAxis = rotAxis / np.linalg.norm(rotAxis)
        self.log.debug("rotAxis Normalized : " + str(rotAxis))
        self.log.debug("Angle : " + str(angle))
        angle = angle * np.pi / 180
        # defining a quaternion for rotatoin
        angle = angle / 2
        rotAxis = rotAxis * np.sin(angle)
        qRot = np.array((np.cos(angle), rotAxis[0], rotAxis[1], rotAxis[2]))
        qRotI = np.array((np.cos(angle), -rotAxis[0], -rotAxis[1], -rotAxis[2]))
        self.log.debug("Rot. quaternion : " + str(qRot))
        self.log.debug("Rot. quaternion conjugate : " + str(qRotI))
        # converting self.kpoints into quaternions
        w = np.zeros((len(self.kpoints), 1))
        qvectors = np.column_stack((w, self.kpoints)).transpose()
        self.log.debug(
            "Kpoints-> quaternions (transposed):\n" + str(qvectors.transpose())
        )
        qvectors = self._q_mult(qRot, qvectors)
        qvectors = self._q_mult(qvectors, qRotI).transpose()
        kpoints = qvectors[:, 1:]
        self.log.debug("Rotated kpoints :\n" + str(qvectors))

        # rotating the spin vector (if exist)
        sxShape, syShape, szShape = self.sx.shape, self.sy.shape, self.sz.shape
        self.log.debug("Spin vector Shapes : " + str((sxShape, syShape, szShape)))
        # The first entry has to be an array of 0s, w could do the work,
        # but if len(self.sx)==0 qvectors will have a non-defined length
        qvectors = (
            0 * self.sx.flatten(),
            self.sx.flatten(),
            self.sy.flatten(),
            self.sz.flatten(),
        )
        self.log.debug("Spin vector quaternions: \n" + str(qvectors))
        qvectors = self._q_mult(qRot, qvectors)
        qvectors = self._q_mult(qvectors, qRotI)
        self.log.debug("Spin quaternions after rotation:\n" + str(qvectors))
        sx, sy, sz = qvectors[1], qvectors[2], qvectors[3]
        sx.shape, sy.shape, sz.shape = sxShape, syShape, szShape

        if store is True:
            self.kpoints, self.sx, self.sy, self.sz = kpoints, sx, sy, sz
        self.log.debug("GeneralRotation: ...Done")
        return (kpoints, sx, sy, sz)

    def RotSymmetryZ(self, order):
        """Applies the given rotational crystal symmetry to the current
    system. ie: to unfold the irreductible BZ to the full BZ.

    Only rotations along z-axis are performed, you can use
    self.GeneralRotation first. 

    The user is responsible of provide a useful input. The method
    doesn't check the physics.

    """
        self.log.debug("RotSymmetryZ:...")
        rotations = [
            self.GeneralRotation(360 * i / order, store=False) for i in range(order)
        ]
        rotations = list(zip(*rotations))
        self.log.debug(
            "self.kpoints.shape (before concat.): " + str(self.kpoints.shape)
        )
        self.kpoints = np.concatenate(rotations[0], axis=0)
        self.log.debug("self.kpoints.shape (after concat.): " + str(self.kpoints.shape))
        self.sx = np.concatenate(rotations[1], axis=0)
        self.sy = np.concatenate(rotations[2], axis=0)
        self.sz = np.concatenate(rotations[3], axis=0)
        # the bands and proj. character also need to be enlarged
        bandsChar = [(self.bands, self.character) for i in range(order)]
        bandsChar = list(zip(*bandsChar))
        self.bands = np.concatenate(bandsChar[0], axis=0)
        self.character = np.concatenate(bandsChar[1], axis=0)
        self.log.debug("RotSymmZ:...Done")

        return

    def MirrorX(self):
        """Applies the given rotational crystal symmetry to the current
    system. ie: to unfold the irreductible BZ to the full BZ.

    """
        self.log.debug("Mirror:...")
        newK = self.kpoints * np.array([1, -1, 1])
        self.kpoints = np.concatenate((self.kpoints, newK), axis=0)
        self.log.debug("self.kpoints.shape (after concat.): " + str(self.kpoints.shape))
        newSx = -1 * self.sx
        newSy = 1 * self.sy
        newSz = 1 * self.sz
        self.sx = np.concatenate((self.sx, newSx), axis=0)
        self.sy = np.concatenate((self.sy, newSy), axis=0)
        self.sz = np.concatenate((self.sz, newSz), axis=0)
        print("self.sx", self.sx.shape)
        print("self.sy", self.sy.shape)
        print("self.sz", self.sz.shape)
        # the bands and proj. character also need to be enlarged
        self.bands = np.concatenate((self.bands, self.bands), axis=0)
        self.character = np.concatenate((self.character, self.character), axis=0)
        print("self.character", self.character.shape)
        print("self.bands", self.bands.shape)
        self.log.debug("Mirror:...Done")

        return

    def Translate(self, newOrigin):
        """Centers the Kpoints at newOrigin, newOrigin is either and index (of
   some Kpoint) or the cartesian coordinates of one point in the
   reciprocal space.

    """
        self.log.debug("Translate():  ...")
        if len(newOrigin) == 1:
            newOrigin = int(newOrigin[0])
            newOrigin = self.kpoints[newOrigin]
        # Make sure newOrigin is a numpy array
        newOrigin = np.array(newOrigin, dtype=float)
        self.log.debug("newOrigin: " + str(newOrigin))
        self.kpoints = self.kpoints - newOrigin
        self.log.debug("new Kpoints:\n" + str(self.kpoints))
        self.log.debug("Translate(): ...Done")
        return



from .utilsprocar import UtilsProcar
from .procarparser import ProcarParser
from .procarselect import ProcarSelect
from .procarplot import ProcarPlot
from .splash import welcome


def Vector(
    infile,
    bands=None,
    energy=None,
    fermi=None,
    atoms=None,
    orbitals=None,
    outcar=None,
    scale=0.1,
):

    welcome()

    print("Input File    : ", infile)
    print("Bands         : ", bands)
    print("Energy        : ", energy)
    print("Fermi         : ", fermi)
    print("outcar        : ", outcar)
    print("atoms         : ", atoms)
    print("orbitals      : ", orbitals)
    print("scale factor  : ", scale)

    if bands is [] and energy is None:
        raise RuntimeError("You must provide the bands or energy.")
    if fermi == None and outcar == None:
        print("WARNING: Fermi's Energy not set")

    # first parse the outcar if given
    recLat = None  # Will contain reciprocal vectors, if necessary
    if outcar:
        outcarparser = UtilsProcar()
        if fermi is None:
            fermi = outcarparser.FermiOutcar(outcar)
            # if quiet is False:
            print("Fermi energy found in outcar file = " + str(fermi))
        recLat = outcarparser.RecLatOutcar(outcar)

    if atoms is None:
        atoms = [-1]
    if orbitals is None:
        orbitals = [-1]

    # parsing the file
    procarFile = ProcarParser()
    procarFile.readFile(infile, recLattice=recLat)

    # processing the data
    sx = ProcarSelect(procarFile, deepCopy=True)
    sx.selectIspin([1])
    sx.selectAtoms(atoms)
    sx.selectOrbital(orbitals)

    sy = ProcarSelect(procarFile, deepCopy=True)
    sy.selectIspin([2])
    sy.selectAtoms(atoms)
    sy.selectOrbital(orbitals)

    sz = ProcarSelect(procarFile, deepCopy=True)
    sz.selectIspin([3])
    sz.selectAtoms(atoms)
    sz.selectOrbital(orbitals)

    x = sx.kpoints[:, 0]
    y = sx.kpoints[:, 1]
    z = sx.kpoints[:, 2]

    # if energy was given I need to find the bands indexes crossing it
    if energy != None:
        FerSurf = FermiSurface(sx.kpoints, sx.bands - fermi, sx.spd, recLat)
        FerSurf.FindEnergy(energy)
        bands = list(FerSurf.useful[0])
        print("Bands indexes crossing Energy  ", energy, ", are: ", bands)

    from mayavi import mlab

    fig = mlab.figure(bgcolor=(1, 1, 1))

    for band in bands:
        # z = sx.bands[:,band]-fermi
        u = sx.spd[:, band]
        v = sy.spd[:, band]
        w = sz.spd[:, band]
        scalar = w

        vect = mlab.quiver3d(
            x,
            y,
            z,
            u,
            v,
            w,
            scale_factor=scale,
            scale_mode="vector",
            scalars=scalar,
            mode="arrow",
            colormap="jet",
        )
        vect.glyph.color_mode = "color_by_scalar"
        vect.scene.parallel_projection = True
        vect.scene.z_plus_view()

        # tube= mlab.plot3d(x,y,z, tube_radius=0.0050, color=(0.5,0.5,0.5))
    mlab.show()



__author__ = "Francisco Munoz,Aldo Romero,Sobhit Singh,Uthpala Herath,Pedram Tavadze,Eric Bousquet,Xu He"
__copyright__ = "Copyright 2020"
__version__ = "5.1.5"
__email__ = (
    "fvmunoz@gmail.com/alromero@mail.wvu.edu/ukh0001@mix.wvu.edu/petavazohi@mix.wvu.edu"
)
__status__ = "Development"
__date__ = "Mar 8th, 2020"

# Copyright (C) 2019 Francisco Munoz,Aldo Romero,Sobhit Singh,Uthpala Herath,Pedram Tavadze,Eric Bousquet,Xu He
#
# This file is part of pyprocar.
#
# Pyprocar is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Phonopy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with pyprocar.  If not, see <http://www.gnu.org/licenses/>.


import sys
import numpy as np
import re
import logging
import matplotlib.pyplot as plt
import seekpath
import ase
import skimage
import pyfiglet

from pyprocar.utilsprocar import UtilsProcar
from pyprocar.procarparser import ProcarParser
from pyprocar.procarplot import ProcarPlot
from pyprocar.procarfilefilter import ProcarFileFilter
from pyprocar.procarselect import ProcarSelect
from pyprocar.fermisurface import FermiSurface
from pyprocar.procarsymmetry import ProcarSymmetry

# ELK
from pyprocar.elkparser import ElkParser

# Abinit
from pyprocar.abinitparser import AbinitParser


from .scriptBandsplot import bandsplot
from .scriptCat import cat
from .scriptFermi2D import fermi2D
from .scriptFermi3D import fermi3D
from .scriptFilter import filter
from .scriptRepair import repair
from .scriptVector import Vector
from .scriptKmesh2D import generate2dkmesh
from .scriptKpath import kpath
from .scriptUnfold import unfold
from .splash import welcome



import pyfiglet
import pyprocar


def welcome():
    print(pyfiglet.figlet_format("PyProcar"))
    print("A Python library for electronic structure pre/post-processing.\n")
    print("Version %s created on %s\n" % (pyprocar.__version__, pyprocar.__date__))
    print(
        "Please cite: Herath, U., Tavadze, P., He, X., Bousquet, E., Singh, S., Muñoz, F. & Romero,\
    A., PyProcar: A Python library for electronic structure pre/post-processing.,\
    Computer Physics Communications 107080 (2019).\n"
    )

    return



import numpy as np
import re
import logging
import matplotlib.pyplot as plt
import sys


class UtilsProcar:
    """
  This class store handy methods that do not fit any other place
  
  members:

  -Openfile: Tries to open a File, it has suitable values for PROCARs
   and can handle gzipped files
   
  -MergeFiles: concatenate two or more PROCAR files taking care of
   metadata and kpoint indexes. Useful for splitted bandstructures
   calculation.

  -FermiOutcar: it greps the Fermi Energy from a given outcar file.

  -RecLatOutcar: it greps the reciprocal lattice from the outcar.

  """

    def __init__(self, loglevel=logging.WARNING):
        self.log = logging.getLogger("UtilsProcar")
        self.log.setLevel(loglevel)
        self.ch = logging.StreamHandler()
        self.ch.setFormatter(
            logging.Formatter("%(name)s::%(levelname)s:" " %(message)s")
        )
        self.ch.setLevel(logging.DEBUG)
        self.log.addHandler(self.ch)
        self.log.debug("UtilsProcar()")
        self.log.debug("UtilsProcar()...done")
        return

    ###############################SCRIPTS####################################################################

    # #calls ProcarRepair
    # def scriptRepair(self,infile,outfile):

    #   print "Input File    : ", infile
    #   print "Output File   : ", outfile

    #   #parsing the file
    #   handler = UtilsProcar()
    #   handler.ProcarRepair(infile,outfile)

    # calls MergeFiles
    # inFiles should be a list of the PROCAR files that require concatenation
    # def scriptCat(self,inFiles,outFile,gz=False):
    #   print   "Concatenating:"
    #   print   "Input         : ", ', '.join(inFiles)
    #   print   "Output        : ", outFile
    #   if gz==True:
    #       print "out compressed: True"

    #   if gz=="True" and outFile[-3:] is not '.gz':
    #     outFile += '.gz'
    #     print ".gz extension appended to the outFile"

    #   handler = UtilsProcar()
    #   handler.MergeFiles(inFiles,outFile, gzipOut=gz)
    #   return

    ####################################################################################################################################################

    def OpenFile(self, FileName=None):
        """
    Tries to open a File, it has suitable values for PROCAR and can
    handle gzipped files

    Example: 

    >>> foo =  UtilsProcar.Openfile()
    Tries to open "PROCAR", then "PROCAR.gz"

    >>> foo = UtilsProcar.Openfile("../bar")
    Tries to open "../bar". If it is a directory, it will try to open
    "../bar/PROCAR" and if fails again "../bar/PROCAR.gz"

    >>> foo = UtilsProcar.Openfile("PROCAR-spd.gz")
    Tries to open a gzipped file "PROCAR-spd.gz"

    If unable to open a file, it raises a "IOError" exception.
"""
        import os
        import gzip

        self.log.debug("OpenFile()")
        self.log.debug("Filename :" + FileName)

        if FileName is None:
            FileName = "PROCAR"
            self.log.debug("Input was None, now is: " + FileName)

        # checking if fileName is just a path and needs a "PROCAR to " be
        # appended
        elif os.path.isdir(FileName):
            self.log.info("The filename is a directory")
            if FileName[-1] != r"/":
                FileName += "/"
            FileName += "PROCAR"
            self.log.debug("I will try  to open :" + FileName)

        # checking that the file exist
        if os.path.isfile(FileName):
            self.log.debug("The File does exist")
            # Checking if compressed
            if FileName[-2:] == "gz":
                self.log.info("A gzipped file found")
                inFile = gzip.open(FileName, mode="rt")
            else:
                self.log.debug("A normal file found")
                inFile = open(FileName, "r")
            return inFile

        # otherwise a gzipped version may exist
        elif os.path.isfile(FileName + ".gz"):
            self.log.info(
                "File not found, however a .gz version does exist and will" " be used"
            )
            inFile = gzip.open(FileName + ".gz", mode="rt")

        else:
            self.log.debug("File not exist, neither a gzipped version")
            print(FileName)
            raise IOError("File not found")

        self.log.debug("OpenFile()...done")
        return inFile

    def MergeFiles(self, inFiles, outFile, gzipOut=False):
        """
    Concatenate two or more PROCAR files. This methods
    takes care of the k-indexes.

    Useful when large number of K points have been calculated in
    different PROCARs.
    
    Args:
    -inFiles: an iterable with files to be concatenated

    -outFile: a string with the outfile name.

    -gzipOut: whether gzip or not the outout file.

    Warning: spin polarized case is not Ok!

    """
        import gzip

        self.log.debug("MergeFiles()")
        self.log.debug("infiles: " " ,".join(inFiles))

        inFiles = [self.OpenFile(x) for x in inFiles]
        header = [x.readline() for x in inFiles]
        self.log.debug("All the input headers are: \n" + "".join(header))
        metas = [x.readline() for x in inFiles]
        self.log.debug("All the input metalines are:\n " + "".join(metas))
        # parsing metalines

        parsedMeta = [list(map(int, re.findall(r"#[^:]+:([^#]+)", x))) for x in metas]
        kpoints = [x[0] for x in parsedMeta]
        bands = set([x[1] for x in parsedMeta])
        ions = set([x[2] for x in parsedMeta])

        # checking that bands and ions match (mind: bands & ions are 'sets'):
        if len(bands) != 1 or len(ions) != 1:
            self.log.error("Number of bands/ions  do not match")
            raise RuntimeError("Files are incompatible")

        newKpoints = np.array(kpoints, dtype=int).sum()
        self.log.info("New number of Kpoints: " + str(newKpoints))
        newMeta = metas[0].replace(str(kpoints[0]), str(newKpoints), 1)
        self.log.debug("New meta line:\n" + newMeta)

        if gzipOut:
            self.log.debug("gzipped output")
            outFile = gzip.open(outFile, mode="wt")
        else:
            self.log.debug("normal output")
            outFile = open(outFile, "w")
        outFile.write(header[0])
        outFile.write(newMeta)

        # embedded function to change old k-point indexes by the correct
        # ones. The `kreplace.k` syntax is for making the variable 'static'
        def kreplace(matchobj):
            # self.log.debug(print matchobj.group(0))
            kreplace.k += 1
            kreplace.localCounter += 1
            return matchobj.group(0).replace(
                str(kreplace.localCounter), str(kreplace.k)
            )

        kreplace.k = 0
        down = []  # to handle spin-down (if found)
        self.log.debug("Going to replace K-points indexes")
        for inFile in inFiles:
            lines = inFile.read()
            # looking for an extra metada line, if found the file is
            # spin-polarized
            p = re.compile(
                r"#[\s\w]+k-points:[\s\d]+#[\s\w]+bands:[\s\d]+#[\w\s]+ions:\s*\d+\s*"
            )
            lines = p.split(lines)
            up = lines[0]
            if len(lines) == 2:
                down.append(lines[1])
                self.log.info("Spin-polarized PROCAR!")
            # closing inFile
            inFile.close()
            kreplace.localCounter = 0
            up = re.sub("(\s+k-point\s*\d+\s*:)", kreplace, up)
            outFile.write(up)

        # handling the spin-down channel, if present
        if down:
            self.log.debug("writing spin down metadata")
            outFile.write("\n")
            outFile.write(newMeta)
        kreplace.k = 0
        for group in down:
            kreplace.localCounter = 0
            group = re.sub("(\s+k-point\s*\d+\s*:)", kreplace, group)
            outFile.write(group)

        self.log.debug("Closing output file")
        outFile.close()
        self.log.debug("MergeFiles()...done")
        return

    def FermiOutcar(self, filename):
        """Just finds all E-fermi fields in the outcar file and keeps the
    last one (if more than one found).

    Args:
    -filename: the file name of the outcar to be readed

    """
        self.log.debug("FermiOutcar(): ...")
        self.log.debug("Input filename : " + filename)

        outcar = open(filename, "r").read()
        match = re.findall(r"E-fermi\s*:\s*(-?\d+.\d+)", outcar)[-1]
        self.log.info("Fermi Energy found : " + match)
        self.log.debug("FermiOutcar(): ...Done")
        return float(match)

    def RecLatOutcar(self, filename):
        """Finds and return the reciprocal lattice vectors, if more than
    one set present, it return just the last one.

    Args: 
    -filename: the name of the outcar file  to be read
    
    """
        self.log.debug("RecLatOutcar(): ...")
        self.log.debug("Input filename : " + filename)

        outcar = open(filename, "r").read()
        # just keeping the last component
        recLat = re.findall(r"reciprocal\s*lattice\s*vectors\s*([-.\s\d]*)", outcar)[-1]
        self.log.debug("the match is : " + recLat)
        recLat = recLat.split()
        recLat = np.array(recLat, dtype=float)
        # up to now I have, both direct and rec. lattices (3+3=6 columns)
        recLat.shape = (3, 6)
        recLat = recLat[:, 3:]
        self.log.info("Reciprocal Lattice found :\n" + str(recLat))
        self.log.debug("RecLatOutcar(): ...Done")
        return recLat

    def ProcarRepair(self, infilename, outfilename):
        """It Tries to repair some stupid problems due the stupid fixed
    format of the stupid fortran.

    Up to now it only separes k-points as the following:
    k-point    61 :    0.00000000-0.50000000 0.00000000 ...
    to
    k-point    61 :    0.00000000 -0.50000000 0.00000000 ...

    But as I found new stupid errors they should be fixed here.
    """
        self.log.debug("ProcarRepair(): ...")
        infile = self.OpenFile(infilename)
        fileStr = infile.read()
        infile.close()

        # Fixing bands issues (when there are more than 999 bands)
        # band *** # energy    6.49554019 # occ.  0.00000000
        fileStr = re.sub(r"(band\s)(\*\*\*)", r"\1 1000", fileStr)

        # Fixing k-point issues
        fileStr = re.sub(r"(\.\d{8})(\d{2}\.)", r"\1 \2", fileStr)
        fileStr = re.sub(r"(\d)-(\d)", r"\1 -\2", fileStr)

        fileStr = re.sub(r"\*+", r" -10.0000 ", fileStr)

        outfile = open(outfilename, "w")
        outfile.write(fileStr)
        outfile.close()

        self.log.debug("ProcarRepair(): ...Done")
        return



import numpy as np
import re
from ase.io import read
from .unfolder import Unfolder
import matplotlib.pyplot as plt
from .fatband import plot_band_weight
from ..procarparser import ProcarParser


class ProcarUnfolder(object):
    def __init__(self, procar, poscar, supercell_matrix, ispin=None):
        self.fname = procar
        self.supercell_matrix = supercell_matrix
        self._parse_procar(ispin=ispin)
        self.atoms = read(poscar)
        self.basis = []
        self.positions = []

    def _parse_procar(self, ispin=None):
        self.procar = ProcarParser()
        self.procar.readFile2(self.fname, phase=True, ispin=ispin)

    def _prepare_unfold_basis(self, ispin=None):
        # basis, which are the name of the bands e.g. 'Ti|dxy|0'
        # self.eigenvectors = np.zeros(
        #    (self.procar.kpointsCount, self.procar.bandsCount,
        #     (self.procar.orbitalCount - 1) * (self.procar.ionsCount - 1) *
        #     self.procar.ispin), dtype='complex')
        if ispin is None:
            iispin = 0
        else:
            ispin -= 1
        self.eigenvectors = np.reshape(
            self.procar.carray[:, :, ispin, :, :],
            (
                self.procar.kpointsCount,
                self.procar.bandsCount,
                self.procar.ionsCount * self.procar.orbitalCount,
            ),
        )
        norm = np.linalg.norm(self.eigenvectors, ord=2, axis=2)
        self.eigenvectors /= norm[:, :, None]

        for iatom, chem in enumerate(self.atoms.get_chemical_symbols()):
            for iorb, orb in enumerate(self.procar.orbitalName):
                for spin in range(self.procar.nspin):
                    # todo: what about spin?
                    self.basis.append("%s|%s|%s" % (None, orb, spin))
                    self.positions.append(self.atoms.get_scaled_positions()[iatom])

    def unfold(self, ispin=None):
        # spd: spd[kpoint][band][ispin][atom][orbital]
        # bands[kpt][iband]
        # to unfold,
        # unfolder:
        # def __init__(self, cell, basis, positions , supercell_matrix, eigenvectors, qpoints, tol_r=0.1, compare=None):
        self._prepare_unfold_basis(ispin=ispin)
        self.unfolder = Unfolder(
            self.atoms.cell,
            self.basis,
            self.positions,
            self.supercell_matrix,
            self.eigenvectors,
            self.procar.kpoints,
            phase=False,
        )
        w = self.unfolder.get_weights()
        return w

    def plot(
        self,
        efermi=5.46,
        ispin=None,
        ylim=(-5, 10),
        ktick=[0, 41, 83, 125, 200],
        kname=["$\Gamma$", "X", "M", "R", "$\Gamma$"],
        show_band=True,
        shift_efermi=True,
        width=4.0,
        color="blue",
        axis=None,
        savetab=None,
    ):
        iispin = 0
        if ispin is not None:
            iispin = ispin - 1
        xlist = [list(range(self.procar.kpointsCount))]
        uf = self.unfold(ispin=ispin)
        if savetab is not None:
            nk, nb = uf.shape
            tab = np.zeros((nb, nk * 2), dtype=float)
            tab[:, ::2] = self.procar.bands[iispin].T
            tab[:, 1::2] = uf.T
            np.savetxt(
                savetab,
                tab,
                delimiter=",",
                fmt="%10.4f",
                header="# nkpoints: %s   nbands:%s \n#E(k1) w(k1) E(k2) w(k2) E(k3) w(k3)..."
                % (nk, nb),
            )
        axes = plot_band_weight(
            xlist * self.procar.bandsCount,
            self.procar.bands[iispin].T,
            np.abs(uf.T),
            xticks=[kname, ktick],
            efermi=efermi,
            shift_efermi=shift_efermi,
            fatness=width,
            color=color,
            axis=axis,
        )
        axes.set_ylim(ylim)
        axes.set_xlim(0, self.procar.kpointsCount - 1)
        if shift_efermi:
            shift = -efermi
        else:
            shift = 0.0
        if show_band:
            for i in range(self.procar.bandsCount):
                axes.plot(
                    self.procar.bands[iispin, :, i] + shift,
                    color="gray",
                    linewidth=1,
                    alpha=0.3,
                )
        return axes



import numpy as np
import matplotlib.pyplot as plt
from .procarunfold import ProcarUnfolder
from .utilsprocar import UtilsProcar
from .splash import welcome


def unfold(
    fname="PROCAR",
    poscar="POSCAR",
    outcar="OUTCAR",
    supercell_matrix=np.diag([2, 2, 2]),
    ispin=None,
    efermi=None,
    shift_efermi=True,
    elimit=(-5, 15),
    kticks=[0, 36, 54, 86, 110, 147, 165, 199],
    knames=["$\Gamma$", "K", "M", "$\Gamma$", "A", "H", "L", "A"],
    print_kpts=False,
    show_band=True,
    width=4,
    color="blue",
    savetab="unfold_result.csv",
    savefig="unfolded_band.png",
    exportplt=False,
):

    """

    Parameters
    ---------- 
    fname: PROCAR filename.
    poscar: POSCAR filename
    outcar: OUTCAR filename, for reading fermi energy. You can also use efermi and set outcar=None
    supercell_matrix: supercell matrix from primitive cell to supercell
    ispin: For non-spin polarized system, ispin=None. 
           For spin polarized system: ispin=1 is spin up, ispin=2 is spin down.
    efermi: Fermi energy
    elimit: range of energy to be plotted.
    kticks: the indices of K points which has labels given in knames.
    knames: see kticks
    print_kpts: print all the kpoints to screen. This is to help find the kticks and knames.
    show_band: whether to plot the bands before unfolding.
    width: the width of the unfolded band. 
    color: color of the unfoled band.
    savetab: the csv file name of which  the table of unfolding result will be written into.
    savefig: the file name of which the figure will be saved.
    exportplt: flag to export plot as matplotlib.pyplot object.

    """
    welcome()

    # Turn interactive plotting off
    plt.ioff()

    if efermi is not None:
        fermi = efermi
    elif outcar is not None:
        outcarparser = UtilsProcar()
        fermi = outcarparser.FermiOutcar(outcar)
    else:
        raise Warning("Fermi energy is not given, neither an OUTCAR contains it.")

    uf = ProcarUnfolder(
        procar=fname, poscar=poscar, supercell_matrix=supercell_matrix, ispin=ispin
    )
    if print_kpts:
        for ik, k in enumerate(uf.procar.kpoints):
            print(ik, k)
    axes = uf.plot(
        efermi=fermi,
        ispin=ispin,
        shift_efermi=shift_efermi,
        ylim=elimit,
        ktick=kticks,
        kname=knames,
        color=color,
        width=width,
        savetab=savetab,
        show_band=show_band,
    )

    if exportplt:
        return plt

    else:
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")
            plt.close()  # Added by Nicholas Pike to close memory issue of looping and creating many figures
        else:
            plt.show()
        return


# if __name__ == '__main__':
#     """
#     An example of how to use
#     """
#     import pyprocar
#     import numpy as np
#     pyprocar.unfold(
#         fname='PROCAR',
#         poscar='POSCAR',
#         outcar='OUTCAR',
#         supercell_matrix=np.diag([2, 2, 2]),
#         efermi=None,
#         shift_efermi=True,
#         ispin=0,
#         elimit=(-5, 15),
#         kticks=[0, 36, 54, 86, 110, 147, 165, 199],
#         knames=['$\Gamma$', 'K', 'M', '$\Gamma$', 'A', 'H', 'L', 'A'],
#         print_kpts=False,
#         show_band=True,
#         savefig='unfolded_band.png')



from .utilsprocar import UtilsProcar
from .splash import welcome
import re
import os
import numpy as np


def cat(inFiles, outFile, gz=False, mergeparallel=False, fixformat=False):
    """
    This module concatenates multiple PROCARs.
    set fixparallel = True for merging PROCARs generated from
    parallel Abinit calculations. 
    """
    welcome()

    print("Concatenating...")
    print("Input         : ", inFiles)  # ', '.join(inFiles)
    print("Output        : ", outFile)

    if mergeparallel == False and fixformat == False:

        if gz == true:
            print("out compressed: true")

        if gz == "true" and outFile[-3:] is not ".gz":
            outFile += ".gz"
            print(".gz extension appended to the outfile")

        handler = utilsprocar()
        handler.mergefiles(inFiles, outFile, gzipout=gz)
        return

    elif mergeparallel == True and fixformat == False:
        _mergeparallel(inFiles, outFile)

    elif mergeparallel == True and fixformat == True:
        outFile_temp = 'outFile.tmp'
        _mergeparallel(inFiles, outFile_temp)
        _fixformat(outFile_temp, outFile)
        if os.path.exists(outFile_temp):
            os.remove(outFile_temp)

    elif mergeparallel == False and fixformat == True:
        print('Using fixformat = True without mergeparallel. Input a single PROCAR.')
        _fixformat(inFiles, outFile)


def _mergeparallel(inputfiles=None, outputfile=None):
    """ This merges Procar files seperated between k-point ranges.
    Happens with parallel Abinit runs. 
    """
    print("Merging parallel files...")
    filenames = sorted(inputfiles)

    with open(outputfile, "w") as outfile:
        for fname in filenames:
            with open(fname) as infile:
                for line in infile:
                    outfile.write(line)


def _fixformat(inputfile=None, outputfile=None):

    """Fixes the formatting of Abinit's Procar
    when the tot projection is not summed and spin directions
    not seperated.
    """
    print("Fixing formatting errors...")
    # removing existing temporary fixed file
    if os.path.exists(outputfile):
        os.remove(outputfile)

    ####### Fixing the parallel PROCARs from Abinit ##########

    rf = open(inputfile, "r")
    data = rf.read()
    rf.close()

    # reading headers
    rffl = open(inputfile, "r")
    first_line = rffl.readline()
    rffl.close()

    # header
    header = re.findall("#\sof\s.*", data)[0]

    # writing to PROCAR
    fp = open(outputfile, "a")
    fp.write(first_line)
    fp.write(str(header) + "\n\n")

    # get all the k-point line headers
    kpoints_raw = re.findall("k-point\s*[0-9]\s*:*.*", data)

    for kpoint_counter in range(len(kpoints_raw)):

        if kpoint_counter == (len(kpoints_raw) - 1):
            # get bands of last k point
            bands_raw = re.findall(
                kpoints_raw[kpoint_counter] + "([a-z0-9\s\n.+#-]*)", data
            )[0]

        else:
            # get bands between k point n and n+1
            bands_raw = re.findall(
                kpoints_raw[kpoint_counter]
                + "([a-z0-9\s\n.+#-]*)"
                + kpoints_raw[kpoint_counter + 1],
                data,
            )[0]

        # get the bands headers for a certain k point
        raw_bands = re.findall("band\s*[0-9]*.*", bands_raw)

        # writing k point header to file
        fp.write(kpoints_raw[kpoint_counter] + "\n\n")

        for band_counter in range(len(raw_bands)):

            if band_counter == (len(raw_bands) - 1):
                # the last band
                single_band = re.findall(
                    raw_bands[band_counter] + "([a-z0-9.+\s\n-]*)", bands_raw
                )[0]

            else:
                # get a single band
                single_band = re.findall(
                    raw_bands[band_counter]
                    + "([a-z0-9.+\s\n-]*)"
                    + raw_bands[band_counter + 1],
                    bands_raw,
                )[0]

            # get the column headers for ion, orbitals and total
            column_header = re.findall("ion\s.*tot", single_band)[0]

            # get number of ions using PROCAR file
            nion_raw = re.findall("#\s*of\s*ions:\s*[0-9]*", data)[0]
            nion = int(nion_raw.split(" ")[-1])

            # the first column of the band. Same as ions
            first_column = []
            for x in single_band.split("\n"):
                if x != "":
                    if x != " ":
                        if x.split()[0] != "ion":
                            first_column.append(x.split()[0])

            # number of spin orientations
            norient = int(len(first_column) / nion)

            # calculate the number of orbital headers (s,p,d etc.)
            for x in single_band.split("\n"):
                if x != "":
                    if x != " ":
                        if x.split()[0] == "ion":
                            norbital = len(x.split()) - 2

            # array to store a single band data as seperate lines
            single_band_lines = []
            for x in single_band.split("\n"):
                if x != "":
                    if x != " ":
                        if x.split()[0] != "ion":
                            single_band_lines.append(x)

            # create empty array to store data (the orbitals + tot)
            bands_orb = np.zeros(shape=(norient, nion, norbital + 1))

            # enter data into bands_orb
            iion = 0
            iorient = 0
            for x in single_band.split("\n"):
                if x != "" and x != " " and x.split()[0] != "ion":
                    iline = x.split()
                    if iion > 1:
                        iion = 0
                        iorient += 1
                    for iorb in range(0, norbital + 1):
                        bands_orb[iorient, iion, iorb] = float(iline[iorb + 1])
                    iion += 1

            # create an array to store the total values
            tot = np.zeros(shape=(norient, norbital + 1))

            # entering data into tot array
            for iorient in range(norient):
                tot[iorient, :] = np.sum(bands_orb[iorient, :, :], axis=0)

            # writing data
            fp.write(raw_bands[band_counter] + "\n\n")
            fp.write(column_header + "\n")

            band_iterator = 0
            total_count = 0
            for orientations_count in range(norient):
                for ions_count in range(nion):
                    fp.write(single_band_lines[band_iterator] + "\n")
                    band_iterator += 1
                fp.write("tot  " + " ".join(map(str, tot[total_count, :])) + "\n\n")
                total_count += 1
    fp.close()



from .utilsprocar import UtilsProcar
from .procarparser import ProcarParser
from .procarselect import ProcarSelect
from .procarplot import ProcarPlot
from .elkparser import ElkParser
from .abinitparser import AbinitParser
from .splash import welcome
import numpy as np
import matplotlib.pyplot as plt
import re


def bandsplot(
    file=None,
    mode="scatter",
    color="blue",
    abinit_output=None,
    spin=0,
    atoms=None,
    orbitals=None,
    fermi=None,
    elimit=None,
    mask=None,
    markersize=0.02,
    cmap="jet",
    vmax=None,
    vmin=None,
    grid=True,
    marker="o",
    permissive=False,
    human=False,
    savefig=None,
    kticks=None,
    knames=None,
    title=None,
    outcar=None,
    kpointsfile=None,
    exportplt=False,
    kdirect=True,
    discontinuities=[],
    code="vasp",
):

    """This function plots band structures
  """
    welcome()

    # Turn interactive plotting off
    plt.ioff()

    # First handling the options, to get feedback to the user and check
    # that the input makes sense.
    # It is quite long
    if atoms is None:
        atoms = [-1]
        if human is True:
            print("WARNING: `--human` option given without atoms list!")
            print("--human will be set to False (ignored)\n ")
            human = False
    if orbitals is None:
        orbitals = [-1]

    print("Script initiated")
    print("code          : ", code)
    print("input file    : ", file)
    print("Mode          : ", mode)
    print("spin comp.    : ", spin)
    print("atoms list   : ", atoms)
    print("orbs. list   : ", orbitals)

    if fermi is None and outcar is None and abinit_output is None and code != "elk":
        print(
            "WARNING: Fermi Energy not set! Please set manually or provide output file."
        )
        fermi = 0

    if fermi is None and code == "elk":
        fermi = None

    print("Fermi Energy   : ", fermi)
    print("Energy range  : ", elimit)

    if mask is not None:
        print("masking thres.: ", mask)

    print("Colormap      : ", cmap)
    print("MarkerSize    : ", markersize)
    print("Permissive    : ", permissive)
    if permissive:
        print("INFO: Permissive flag is on! Be careful")
    print("vmax          : ", vmax)
    print("vmin          : ", vmin)
    print("grid enabled  : ", grid)
    if human is not None:
        print("human         : ", human)
    print("Savefig       : ", savefig)
    print("title         : ", title)
    print("outcar        : ", outcar)

    if kdirect:
        print("k-point coordinates        : reduced")
    else:
        print(
            "k-point coordinates        : cartesian (Remember to supply an output file for this case to work.)"
        )

    #### READING KPOINTS FILE IF PRESENT ####

    # If KPOINTS file is given:
    if kpointsfile is not None:
        # Getting the high symmetry point names from KPOINTS file
        f = open(kpointsfile)
        KPread = f.read()
        f.close()

        KPmatrix = re.findall("reciprocal[\s\S]*", KPread)
        tick_labels = np.array(re.findall("!\s(.*)", KPmatrix[0]))
        knames = []
        knames = [tick_labels[0]]

        ################## Checking for discontinuities ########################
        discont_indx = []
        icounter = 1
        while icounter < len(tick_labels) - 1:
            if tick_labels[icounter] == tick_labels[icounter + 1]:
                knames.append(tick_labels[icounter])
                icounter = icounter + 2
            else:
                discont_indx.append(icounter)
                knames.append(tick_labels[icounter] + "/" + tick_labels[icounter + 1])
                icounter = icounter + 2
        knames.append(tick_labels[-1])
        discont_indx = list(dict.fromkeys(discont_indx))

        ################# End of discontinuity check ##########################

        # Added by Nicholas Pike to modify the output of seekpath to allow for
        # latex rendering.
        for i in range(len(knames)):
            if knames[i] == "GAMMA":
                knames[i] = "\Gamma"
            else:
                pass

        knames = [str("$" + latx + "$") for latx in knames]

        # getting the number of grid points from the KPOINTS file
        f2 = open(kpointsfile)
        KPreadlines = f2.readlines()
        f2.close()
        numgridpoints = int(KPreadlines[1].split()[0])

        kticks = [0]
        gridpoint = 0
        for kt in range(len(knames) - 1):
            gridpoint = gridpoint + numgridpoints
            kticks.append(gridpoint - 1)

        print("knames        : ", knames)
        print("kticks        : ", kticks)

        # creating an array for discontunuity k-points. These are the indexes
        # of the discontinuity k-points.
        discontinuities = []
        for k in discont_indx:
            discontinuities.append(kticks[int(k / 2) + 1])
        if discontinuities:
            print("discontinuities :", discontinuities)

    #### END OF KPOINTS FILE DEPENDENT SECTION ####

    # The spin argument should be a number (index of an array), or
    #'st'. In the last case it will be handled separately (later)

    spin = {"0": 0, "1": 1, "2": 2, "3": 3, "st": "st"}[str(spin)]

    #### parsing the PROCAR file or equivalent to retrieve spd data ####

    if code == "vasp":
        procarFile = ProcarParser()

    elif code == "abinit":
        procarFile = ProcarParser()
        abinitFile = AbinitParser(abinit_output=abinit_output)

    elif code == "elk":
        # reciprocal lattice already taken care of
        procarFile = ElkParser(kdirect=kdirect)

        # Retrieving knames and kticks from Elk
        if kticks is None and knames is None:
            if procarFile.kticks and procarFile.knames:
                kticks = procarFile.kticks
                knames = procarFile.knames

    # If ticks and names are given by the user manually:
    if kticks is not None and knames is not None:
        ticks = list(zip(kticks, knames))
    elif kticks is not None:
        ticks = list(zip(kticks, kticks))
    else:
        ticks = None

    if kpointsfile is None and kticks and knames:
        print("kticks        : ", kticks)
        print("knames        : ", knames)
        if discontinuities:
            print("discontinuities :", discontinuities)

    # The second part of this function is parse/select/use the data in
    # OUTCAR (if given) and PROCAR

    # first parse the outcar if given, to get Efermi and Reciprocal lattice
    recLat = None
    if code == "vasp":
        if outcar:
            outcarparser = UtilsProcar()
            if fermi is None:
                fermi = outcarparser.FermiOutcar(outcar)
                print("Fermi energy found in OUTCAR file = " + str(fermi))
            recLat = outcarparser.RecLatOutcar(outcar)

    elif code == "elk":
        if fermi is None:
            fermi = procarFile.fermi
            print("Fermi energy found in Elk output file = " + str(fermi))

    elif code == "abinit":
        if fermi is None:
            fermi = abinitFile.fermi
            print("Fermi energy found in Abinit output file = " + str(fermi))
        recLat = abinitFile.reclat

    # if kdirect = False, then the k-points will be in cartesian coordinates.
    # The output should be read to find the reciprocal lattice vectors to transform
    # from direct to cartesian

    if code == "vasp":
        if kdirect:
            procarFile.readFile(file, permissive)
        else:
            procarFile.readFile(file, permissive, recLattice=recLat)

    elif code == 'abinit':
        if kdirect:
            procarFile.readFile(file, permissive)
        else:
            procarFile.readFile(file, permissive, recLattice=recLat)

    # processing the data, getting an instance of the class that reduces the data
    data = ProcarSelect(procarFile, deepCopy=True)

    # handling the spin, `spin='st'` is not straightforward, needs
    # to calculate the k vector and its normal. Other `spin` values
    # are trivial.
    if spin is "st":
        # two `ProcarSelect` instances, to store temporal values: spin_x, spin_y
        dataX = ProcarSelect(procarFile, deepCopy=True)
        dataX.selectIspin([1])
        dataX.selectAtoms(atoms, fortran=human)
        dataX.selectOrbital(orbitals)
        dataY = ProcarSelect(procarFile, deepCopy=True)
        dataY.selectIspin([2])
        dataY.selectAtoms(atoms, fortran=human)
        dataY.selectOrbital(orbitals)
        # getting the signed angle of each K-vector
        angle = np.arctan2(dataX.kpoints[:, 1], (dataX.kpoints[:, 0] + 0.000000001))
        sin = np.sin(angle)
        cos = np.cos(angle)
        sin.shape = (sin.shape[0], 1)
        cos.shape = (cos.shape[0], 1)
        ##print sin, cos
        # storing the spin projection into the original array
        data.spd = -sin * dataX.spd + cos * dataY.spd
    else:
        data.selectIspin([spin])
        data.selectAtoms(atoms, fortran=human)
        data.selectOrbital(orbitals)

    # Plotting the data
    data.bands = (data.bands.transpose() - np.array(fermi)).transpose()
    plot = ProcarPlot(data.bands, data.spd, data.kpoints)

    ###### start of mode dependent options #########

    if mode == "scatter":
        plot.scatterPlot(
            mask=mask,
            size=markersize,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            marker=marker,
            ticks=ticks,
            discontinuities=discontinuities,
        )
        if fermi is not None:
            plt.ylabel(r"$E-E_f$ [eV]", fontsize=22)
        else:
            plt.ylabel(r"Energy [eV]", fontsize=22)
        if elimit is not None:
            plt.ylim(elimit)

    elif mode == "plain":
        plot.plotBands(
            markersize,
            marker=marker,
            ticks=ticks,
            color=color,
            discontinuities=discontinuities,
        )
        if fermi is not None:
            plt.ylabel(r"$E-E_f$ [eV]", fontsize=22)
        else:
            plt.ylabel(r"Energy [eV]", fontsize=22)
        if elimit:
            plt.ylim(elimit)

    elif mode == "parametric":
        plot.parametricPlot(
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            ticks=ticks,
            discontinuities=discontinuities,
        )
        if fermi is not None:
            plt.ylabel(r"$E-E_f$ [eV]", fontsize=22)
        else:
            plt.ylabel(r"Energy [eV]", fontsize=22)
        if elimit is not None:
            plt.ylim(elimit)

    elif mode == "atomic":
        plot.atomicPlot(cmap=cmap, vmin=vmin, vmax=vmax)
        if fermi is not None:
            plt.ylabel(r"$E-E_f$ [eV]", fontsize=22)
        else:
            plt.ylabel(r"Energy [eV]", fontsize=22)
        if elimit is not None:
            plt.ylim(elimit)
    ###### end of mode dependent options ###########
    plt.tight_layout()
    if grid:
        plt.grid()

    if title:
        plt.title(title, fontsize=22)

    if exportplt:
        return plt

    else:
        if savefig:
            plt.savefig(savefig, bbox_inches="tight")
            plt.close()  # Added by Nicholas Pike to close memory issue of looping and creating many figures
        else:
            plt.show()
        return


# if __name__ == "__main__":
# bandsplot(mode='parametric',elimit=[-6,6],orbitals=[4,5,6,7,8],vmin=0,vmax=1, code='elk')
# knames=['$\Gamma$', '$X$', '$M$', '$\Gamma$', '$R$','$X$'],
# kticks=[0, 8, 16, 24, 38, 49])



import re
import os
import seekpath
import numpy as np


def kpath(
    infile,
    grid_size,
    with_time_reversal,
    recipe,
    threshold,
    symprec,
    angle_tolerence,
    supercell_matrix=np.eye(3),
):
    """
	This module creates a KPOINTS file for band structure plotting.
	"""
    welcome()

    file = open(infile, "r")
    POSCAR = file.readlines()

    # cell
    cell_matrix = POSCAR[2:5]
    cell = np.zeros(shape=(3, 3))

    for i in range(len(cell_matrix)):
        cell_matrix0 = np.array(cell_matrix[i].split())
        cell[i, :] = (cell_matrix0.astype(np.float)) * np.array(
            POSCAR[1].split()
        ).astype(np.float)

    # positions
    # POSCAR index changed by Nicholas Pike from 5 -> 6 and from 7 -> 8
    # Previously, POSCAR[5] referenced the atom names i.e. Na Cl and not the
    # atom numbers
    atoms = np.array(POSCAR[6].split()).astype(np.int)
    positions_matrix = POSCAR[8 : 8 + sum(atoms)]
    positions = np.zeros(shape=(np.sum(atoms), 3))

    for j in range(len(positions_matrix)):
        positions_matrix0 = np.array(positions_matrix[j].split())[0:3]
        positions[j, :] = positions_matrix0.astype(np.float)

    # numbers
    numbers = np.zeros(sum(atoms))
    counter = 0
    atom_counter = 1

    for ii in atoms:
        for kk in range(ii):
            numbers[counter] = atom_counter
            counter = counter + 1
        atom_counter = atom_counter + 1

    # seekpath
    structure = (cell, positions, numbers)
    kpath_dictionary = seekpath.get_path(
        structure, with_time_reversal, recipe, threshold, symprec, angle_tolerence
    )

    path_array = [""] * 2 * len(kpath_dictionary["path"])
    count = 0
    count2 = 1
    for path_counter in kpath_dictionary["path"]:
        path_array[count] = path_counter[0]
        path_array[count2] = path_counter[1]
        count = count + 2
        count2 = count2 + 2

    coord_matrix = np.zeros(shape=(2 * len(kpath_dictionary["path"]), 3))
    path_array_counter = 0
    for mm in range(len(coord_matrix)):
        coord_matrix[mm, :] = np.dot(
            kpath_dictionary["point_coords"][path_array[path_array_counter]],
            supercell_matrix,
        )
        path_array_counter = path_array_counter + 1

    k_file = open("KPOINTS", "w+")
    k_file.write("KPOINTS generated by PyProcar\n")
    k_file.write("%d ! Grid points\n" % grid_size)
    k_file.write("Line_mode\n")
    k_file.write("reciprocal\n")
    for iterator in range(len(coord_matrix)):
        if iterator % 2 == 0:
            k_file.write(
                "%f %f %f ! %s\n"
                % (
                    coord_matrix[iterator, 0],
                    coord_matrix[iterator, 1],
                    coord_matrix[iterator, 2],
                    path_array[iterator],
                )
            )
        else:
            k_file.write(
                "%f %f %f ! %s\n\n"
                % (
                    coord_matrix[iterator, 0],
                    coord_matrix[iterator, 1],
                    coord_matrix[iterator, 2],
                    path_array[iterator],
                )
            )
    k_file.close()
